package controllers_test

import (
	"context"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	kube_core "k8s.io/api/core/v1"
	kube_meta "k8s.io/apimachinery/pkg/apis/meta/v1"
	kube_types "k8s.io/apimachinery/pkg/types"
	kube_ctrl "sigs.k8s.io/controller-runtime"
	kube_client "sigs.k8s.io/controller-runtime/pkg/client"
	kube_client_fake "sigs.k8s.io/controller-runtime/pkg/client/fake"
	kube_reconcile "sigs.k8s.io/controller-runtime/pkg/reconcile"

	"github.com/kumahq/kuma/pkg/core"
	"github.com/kumahq/kuma/pkg/plugins/runtime/k8s/controllers"
	"github.com/kumahq/kuma/pkg/plugins/runtime/k8s/metadata"

	v1 "github.com/kumahq/kuma/pkg/plugins/runtime/k8s/apis/k8s.cni.cncf.io/v1"
)

var _ = Describe("NamespaceReconciler", func() {

	var kubeClient kube_client.Client
	var reconciler kube_reconcile.Reconciler

	BeforeEach(func() {
		kubeClient = kube_client_fake.NewFakeClientWithScheme(
			k8sClientScheme,
			&kube_core.Namespace{
				ObjectMeta: kube_meta.ObjectMeta{
					Name:      "non-system-ns-with-sidecar-injection",
					Namespace: "non-system-ns-with-sidecar-injection",
					Annotations: map[string]string{
						"kuma.io/sidecar-injection": "enabled",
					},
				},
			},
			&kube_core.Namespace{
				ObjectMeta: kube_meta.ObjectMeta{
					Name:      "non-system-ns-without-sidecar-injection",
					Namespace: "non-system-ns-without-sidecar-injection",
				},
			},
		)

		reconciler = &controllers.NamespaceReconciler{
			Client:     kubeClient,
			CNIEnabled: true,
			Log:        core.Log.WithName("test"),
		}
	})

	It("should create NetworkAttachmentDefinition", func() {
		// given
		req := kube_ctrl.Request{
			NamespacedName: kube_types.NamespacedName{
				Namespace: "non-system-ns-with-sidecar-injection",
				Name:      "non-system-ns-with-sidecar-injection",
			},
		}

		// when
		result, err := reconciler.Reconcile(req)

		// then
		Expect(err).ToNot(HaveOccurred())
		Expect(result).To(BeZero())

		// and NetworkAttachmentDefinition is created
		nads := &v1.NetworkAttachmentDefinitionList{}
		err = kubeClient.List(context.Background(), nads)
		Expect(err).ToNot(HaveOccurred())
		Expect(nads.Items).To(HaveLen(1))
		Expect(nads.Items[0].Namespace).To(Equal("non-system-ns-with-sidecar-injection"))
		Expect(nads.Items[0].Name).To(Equal("kuma-cni"))
	})

	It("should delete NetworkAttachmentDefinition when injection annotation is no longer on the namespace", func() {
		// setup NetworkAttachmentDefinition in the namespace
		nad := &v1.NetworkAttachmentDefinition{
			ObjectMeta: kube_meta.ObjectMeta{
				Namespace: "non-system-ns-without-sidecar-injection",
				Name:      metadata.KumaCNI,
			},
		}
		err := kubeClient.Create(context.Background(), nad)
		Expect(err).ToNot(HaveOccurred())

		// given namespace without kuma.io/sidecar-injection annotation
		req := kube_ctrl.Request{
			NamespacedName: kube_types.NamespacedName{
				Namespace: "non-system-ns-without-sidecar-injection",
				Name:      "non-system-ns-without-sidecar-injection",
			},
		}

		// when
		result, err := reconciler.Reconcile(req)

		// then
		Expect(err).ToNot(HaveOccurred())
		Expect(result).To(BeZero())

		// and NetworkAttachmentDefinition is deleted
		nads := &v1.NetworkAttachmentDefinitionList{}
		err = kubeClient.List(context.Background(), nads)
		Expect(err).ToNot(HaveOccurred())
		Expect(nads.Items).To(HaveLen(0))
	})

	It("should ignore namespace namespaces without label", func() {
		// given
		req := kube_ctrl.Request{
			NamespacedName: kube_types.NamespacedName{
				Namespace: "non-system-ns-without-sidecar-injection",
				Name:      "non-system-ns-without-sidecar-injection",
			},
		}

		// when
		result, err := reconciler.Reconcile(req)

		// then
		Expect(err).ToNot(HaveOccurred())
		Expect(result).To(BeZero())

		// and NetworkAttachmentDefinition is not created
		nads := &v1.NetworkAttachmentDefinitionList{}
		err = kubeClient.List(context.Background(), nads)
		Expect(err).ToNot(HaveOccurred())
		Expect(nads.Items).To(HaveLen(0))
	})

})
